# Generated by Django 5.2.4 on 2025-07-27 10:45

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('test_app', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='BookCategory',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('is_removed', models.BooleanField(default=False)),
                ('removed_at', models.DateTimeField(blank=True, null=True)),
                ('name', models.CharField(max_length=100, unique=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='AuthorProfile',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('is_removed', models.BooleanField(default=False)),
                ('removed_at', models.DateTimeField(blank=True, null=True)),
                ('bio', models.TextField(blank=True)),
                ('author', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='profile', to='test_app.author')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Book',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('is_removed', models.BooleanField(default=False)),
                ('removed_at', models.DateTimeField(blank=True, null=True)),
                ('title', models.CharField(max_length=100)),
                ('author', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='test_app.author')),
                ('categories', models.ManyToManyField(blank=True, related_name='books', to='test_app.bookcategory')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='BookMeta',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('is_removed', models.BooleanField(default=False)),
                ('removed_at', models.DateTimeField(blank=True, null=True)),
                ('format', models.CharField(choices=[('paperback', 'Paperback'), ('hardcover', 'Hardcover'), ('ebook', 'eBook')], default='paperback', max_length=30)),
                ('book', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='meta', to='test_app.book')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



from __future__ import annotations

from datetime import datetime
from datetime import timedelta

from django.db import models
from django.db import router
from django.utils import timezone
from django.utils.functional import classproperty

from django_resurrected.constants import SOFT_DELETE_RETENTION_DAYS

from .collectors import Collector
from .managers import ActiveObjectsManager
from .managers import AllObjectsManager
from .managers import RemovedObjectsManager


class SoftDeleteModel(models.Model):
    retention_days: int | None = SOFT_DELETE_RETENTION_DAYS

    is_removed = models.BooleanField(default=False)
    removed_at = models.DateTimeField(null=True, blank=True)

    objects = AllObjectsManager()
    active_objects = ActiveObjectsManager()
    removed_objects = RemovedObjectsManager()

    class Meta:
        abstract = True

    @classproperty
    def retention_limit(cls) -> datetime | None:
        if cls.retention_days is None:
            return None

        return timezone.now() - timedelta(days=cls.retention_days)

    @property
    def is_expired(self) -> bool:
        if self.retention_limit is None:
            return False

        return bool(
            self.is_removed
            and self.removed_at
            and self.removed_at < self.retention_limit,
        )

    def _get_collector(self, using: str | None = None) -> Collector:
        using = using or router.db_for_write(self.__class__, instance=self)
        return Collector(using=using, origin=self)

    def remove(
        self,
        using: str | None = None,
        keep_parents: bool = False,
    ) -> tuple[int, dict[str, int]]:
        collector = self._get_collector(using)
        collector.collect_reverse_related([self], keep_parents=keep_parents)
        return collector.remove()

    def hard_delete(
        self,
        using: str | None = None,
        keep_parents: bool = False,
    ) -> tuple[int, dict[str, int]]:
        return super().delete(using=using, keep_parents=keep_parents)

    def delete(
        self,
        using: str | None = None,
        keep_parents: bool = False,
    ) -> tuple[int, dict[str, int]]:
        if self.is_expired:
            return self.hard_delete(using=using, keep_parents=keep_parents)

        return self.remove(using=using, keep_parents=keep_parents)

    def restore(
        self,
        with_related: bool = False,
        using: str | None = None,
    ) -> tuple[int, dict[str, int]]:
        using = using or router.db_for_write(self.__class__, instance=self)
        collector = self._get_collector(using)
        collector.collect_forward_related([self])
        collector.collect_reverse_related([self], collect_related=with_related)
        return collector.restore()



from __future__ import annotations

from datetime import datetime

import pytest
from model_bakery import baker
from test_app import models


def assert_is_active(*objs):
    for obj in objs:
        obj.refresh_from_db()
        assert obj.is_removed is False
        assert obj.removed_at is None


def assert_is_removed(*objs, removed_at: datetime | None = None):
    for obj in objs:
        obj.refresh_from_db()
        assert obj.is_removed
        assert obj.removed_at if removed_at is None else obj.removed_at == removed_at


def remove_objs(*objs):
    for obj in objs:
        obj.remove()
    assert_is_removed(*objs)


@pytest.fixture
def make_author():
    return lambda **kwargs: baker.make(models.Author, **kwargs)


@pytest.fixture
def make_author_profile():
    return lambda author, **kwargs: baker.make(
        models.AuthorProfile, author=author, **kwargs
    )


@pytest.fixture
def make_book():
    return lambda author, **kwargs: baker.make(models.Book, author=author, **kwargs)


@pytest.fixture
def make_book_meta():
    return lambda book, **kwargs: baker.make(models.BookMeta, book=book, **kwargs)


@pytest.fixture
def make_book_category():
    return lambda **kwargs: baker.make(models.BookCategory, **kwargs)


@pytest.fixture
def test_author(make_author):
    return make_author()


@pytest.fixture
def author_with_profile(make_author, make_author_profile):
    author = make_author()
    profile = make_author_profile(author=author)
    return author, profile


@pytest.fixture
def author_with_books(make_author, make_book):
    author = make_author()
    books = make_book(author=author, _quantity=2)
    return author, books


@pytest.fixture
def book_with_category(make_author, make_book, make_book_category):
    book = make_book(author=make_author())
    category = make_book_category()
    book.categories.add(category)
    return book, category


@pytest.fixture
def make_author_with_all_relations(
    make_author, make_author_profile, make_book, make_book_meta, make_book_category
):
    def _make_author_with_all_relations():
        author = make_author()
        profile = make_author_profile(author=author)
        book_1, book_2 = make_book(author=author, _quantity=2)
        book_meta_1 = make_book_meta(book_1)
        book_meta_2 = make_book_meta(book_2)
        category_1, category_2 = make_book_category(_quantity=2)
        book_1.categories.add(category_1)
        book_2.categories.add(category_2)
        return author, (
            profile,
            book_1,
            book_2,
            book_meta_1,
            book_meta_2,
            category_1,
            category_2,
        )

    return _make_author_with_all_relations



# Generated by Django 5.2.4 on 2025-08-03 19:40

import uuid
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('test_app', '0002_bookcategory_authorprofile_book_bookmeta'),
    ]

    operations = [
        migrations.AlterField(
            model_name='author',
            name='id',
            field=models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name='authorprofile',
            name='id',
            field=models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name='book',
            name='id',
            field=models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name='bookcategory',
            name='id',
            field=models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name='bookmeta',
            name='id',
            field=models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False),
        ),
    ]



from __future__ import annotations

import inspect
from collections import Counter
from collections.abc import Iterable
from collections.abc import Iterator
from typing import TYPE_CHECKING

from django.contrib.admin.utils import NestedObjects
from django.db import models
from django.db.models.options import Options
from django.utils import timezone

if TYPE_CHECKING:
    from django_resurrected.managers import AllObjectsQuerySet
    from django_resurrected.models import SoftDeleteModel


def is_soft_delete(obj: models.Model | type[models.Model]) -> bool:
    from .models import SoftDeleteModel  # noqa: PLC0415

    model_class = obj if inspect.isclass(obj) else type(obj)
    return issubclass(model_class, SoftDeleteModel)


def get_candidate_relations_to_restore(opts: Options) -> Iterator[models.Field]:
    return (
        f
        for f in opts.get_fields(include_hidden=True)
        if not f.auto_created and f.concrete and (f.one_to_one or f.many_to_one)
    )


class Collector(NestedObjects):
    def collect_forward_related(self, objs: Iterable[models.Model]) -> None:
        for obj in objs:
            model = obj.__class__
            self.model_objs[model].add(obj)

            for field in get_candidate_relations_to_restore(model._meta):
                if field.null:
                    continue
                if related_obj := getattr(obj, field.name):
                    self.collect_forward_related([related_obj])

    def collect_reverse_related(self, objs: Iterable[models.Model], **kwargs) -> None:
        return self.collect(objs, **kwargs)

    @property
    def model_objs_for_update(
        self,
    ) -> dict[type[SoftDeleteModel], set[SoftDeleteModel]]:
        return {
            model: objs
            for model, objs in self.model_objs.items()
            if is_soft_delete(model)
        }

    @property
    def querysets_for_update(self) -> list[AllObjectsQuerySet]:
        querysets = []

        for model, objs in self.model_objs_for_update.items():
            if pk_list := [obj.pk for obj in objs if obj.pk is not None]:
                querysets.append(model.objects.filter(pk__in=pk_list))

        return querysets

    def update(self, **kwargs) -> tuple[int, dict[str, int]]:
        counter: Counter[str] = Counter()

        for queryset in self.querysets_for_update:
            count = queryset.update(**kwargs)
            counter[queryset.model._meta.label] += count

        return sum(counter.values()), dict(counter)

    def restore(self) -> tuple[int, dict[str, int]]:
        return self.update(is_removed=False, removed_at=None)

    def remove(self) -> tuple[int, dict[str, int]]:
        return self.update(is_removed=True, removed_at=timezone.now())



from datetime import datetime
from unittest.mock import patch

import pytest
import pytz
from freezegun.api import freeze_time
from test_app.models import Author
from test_app.models import AuthorProfile
from test_app.models import Book
from test_app.models import BookCategory
from test_app.models import BookMeta

from django_resurrected.managers import ActiveObjectsQuerySet
from django_resurrected.managers import AllObjectsQuerySet
from django_resurrected.managers import RemovedObjectsQuerySet
from tests.conftest import assert_is_active
from tests.conftest import assert_is_removed
from tests.conftest import remove_objs


@pytest.mark.django_db
class TestActiveObjectsQuerySet:
    @freeze_time("2025-05-01")
    def test_remove(self, make_author):
        authors = make_author(_quantity=3)
        assert_is_active(*authors)

        Author.active_objects.all().remove()

        assert_is_removed(*authors, removed_at=datetime(2025, 5, 1, tzinfo=pytz.utc))

    @freeze_time("2025-05-01")
    def test_remove_with_related_o2o_cascade(self, make_author, make_author_profile):
        author_1, author_2, author_3 = make_author(_quantity=3)
        profile_1 = make_author_profile(author=author_1)
        profile_2 = make_author_profile(author=author_2)
        assert_is_active(author_1, author_2, author_3, profile_1, profile_2)

        Author.active_objects.filter(id=author_1.id).remove()

        assert_is_active(author_2, author_3, profile_2)
        assert_is_removed(
            author_1, profile_1, removed_at=datetime(2025, 5, 1, tzinfo=pytz.utc)
        )

    @freeze_time("2025-05-01")
    def test_remove_with_related_m2o_cascade(self, make_author, make_book):
        author_1, author_2, author_3 = make_author(_quantity=3)
        book_1 = make_book(author=author_1)
        book_2 = make_book(author=author_2)
        assert_is_active(author_1, author_2, author_3, book_1, book_2)

        Author.active_objects.filter(id=author_1.id).remove()

        assert_is_active(author_2, author_3, book_2)
        assert_is_removed(
            author_1, book_1, removed_at=datetime(2025, 5, 1, tzinfo=pytz.utc)
        )

    @freeze_time("2025-05-01")
    def test_remove_with_related_m2m(self, make_author, book_with_category):
        book, category = book_with_category
        author_1 = book.author
        author_2, author_3 = make_author(_quantity=2)
        assert_is_active(author_1, author_2, author_3, book, category)

        Author.active_objects.filter(id=author_1.id).remove()

        assert_is_active(author_2, author_3, category)
        # NOTE: M2M relations are not removed, as Django doesn't include them in cascade
        # operations.
        assert_is_removed(
            author_1, book, removed_at=datetime(2025, 5, 1, tzinfo=pytz.utc)
        )

    @patch.object(ActiveObjectsQuerySet, "remove")
    def test_delete(self, remove_mock):
        Author.active_objects.all().delete()
        remove_mock.assert_called_once()


@pytest.mark.django_db
class TestActiveObjectsQuerySetWithAllRelations:
    def test_remove_author(self, make_author_with_all_relations):
        author_1, author_1_rels = make_author_with_all_relations()
        author_2, author_2_rels = make_author_with_all_relations()
        profile, book_1, book_2, book_meta_1, book_meta_2, category_1, category_2 = (
            author_1_rels
        )
        Author.active_objects.filter(id=author_1.id).remove()

        assert_is_removed(author_1, profile, book_1, book_2, book_meta_1, book_meta_2)
        assert_is_active(category_1, category_2, author_2, *author_2_rels)

    def test_remove_profile(self, make_author_with_all_relations):
        author_1, author_1_rels = make_author_with_all_relations()
        author_2, author_2_rels = make_author_with_all_relations()
        profile, book_1, book_2, book_meta_1, book_meta_2, category_1, category_2 = (
            author_1_rels
        )

        AuthorProfile.active_objects.filter(id=profile.id).remove()

        assert_is_removed(profile)
        assert_is_active(
            author_1,
            book_1,
            book_2,
            book_meta_1,
            book_meta_2,
            category_1,
            category_2,
            author_2,
            *author_2_rels,
        )

    def test_remove_book(self, make_author_with_all_relations):
        author_1, author_1_rels = make_author_with_all_relations()
        author_2, author_2_rels = make_author_with_all_relations()
        author_3, author_3_rels = make_author_with_all_relations()
        profile, book_1, book_2, book_meta_1, book_meta_2, category_1, category_2 = (
            author_1_rels
        )

        Book.active_objects.filter(id=book_1.id).remove()

        assert_is_removed(book_1, book_meta_1)
        assert_is_active(
            profile,
            book_2,
            book_meta_2,
            category_1,
            category_2,
            author_2,
            *author_2_rels,
        )

    def test_remove_book_meta(self, make_author_with_all_relations):
        author_1, author_1_rels = make_author_with_all_relations()
        author_2, author_2_rels = make_author_with_all_relations()
        author_3, author_3_rels = make_author_with_all_relations()
        profile, book_1, book_2, book_meta_1, book_meta_2, category_1, category_2 = (
            author_1_rels
        )

        BookMeta.active_objects.filter(id=book_meta_1.id).remove()

        assert_is_removed(book_meta_1)
        assert_is_active(
            author_1,
            book_1,
            profile,
            book_2,
            book_meta_2,
            category_1,
            category_2,
            author_2,
            *author_2_rels,
        )

    def test_remove_category(self, make_author_with_all_relations):
        author_1, author_1_rels = make_author_with_all_relations()
        author_2, author_2_rels = make_author_with_all_relations()
        profile, book_1, book_2, book_meta_1, book_meta_2, category_1, category_2 = (
            author_1_rels
        )

        BookCategory.active_objects.filter(id=category_1.id).remove()

        assert_is_removed(category_1)
        assert_is_active(
            author_1,
            profile,
            book_1,
            book_2,
            book_meta_1,
            book_meta_2,
            category_2,
            author_2,
            *author_2_rels,
        )


@pytest.mark.django_db
class TestRemovedObjectsQuerySet:
    def test_restore(self, make_author):
        author_1, author_2, author_3 = make_author(_quantity=3)
        Author.objects.all().remove()
        assert_is_removed(author_1, author_2, author_3)

        Author.removed_objects.filter(id__in=[author_1.id, author_2.id]).restore()

        assert_is_active(author_1, author_2)
        assert_is_removed(author_3)

    def test_restore_without_related(self, make_author, make_author_profile):
        author_1, author_2, author_3 = make_author(_quantity=3)
        profile_1 = make_author_profile(author=author_1)
        profile_2 = make_author_profile(author=author_2)
        Author.objects.all().remove()
        assert_is_removed(author_1, author_2, author_3, profile_1, profile_2)

        result = Author.removed_objects.filter(id=author_1.id).restore()

        assert_is_active(author_1)
        assert_is_removed(author_2, author_3, profile_1, profile_2)
        assert result == (1, {"test_app.Author": 1})

    def test_restore_with_related_o2o_cascade(self, make_author, make_author_profile):
        author_1, author_2, author_3 = make_author(_quantity=3)
        profile_1 = make_author_profile(author=author_1)
        profile_2 = make_author_profile(author=author_2)
        Author.objects.all().remove()
        assert_is_removed(author_1, author_2, author_3, profile_1, profile_2)

        result = Author.removed_objects.filter(id=author_1.id).restore(
            with_related=True
        )

        assert_is_active(author_1, profile_1)
        assert_is_removed(author_2, author_3, profile_2)
        assert result == (2, {"test_app.Author": 1, "test_app.AuthorProfile": 1})

    def test_restore_with_related_m2o_cascade(self, make_author, make_book):
        author_1, author_2, author_3 = make_author(_quantity=3)
        book_1 = make_book(author=author_1)
        book_2 = make_book(author=author_2)
        Author.active_objects.all().remove()
        assert_is_removed(author_1, author_2, author_3, book_1, book_2)

        Author.removed_objects.filter(id=author_1.id).restore(with_related=True)

        assert_is_active(author_1, book_1)
        assert_is_removed(author_2, author_3, book_2)

    def test_restore_with_related_o2m(self, make_author, make_book):
        author_1, author_2, author_3 = make_author(_quantity=3)
        book_1 = make_book(author=author_1)
        book_2 = make_book(author=author_2)
        Author.active_objects.all().remove()
        assert_is_removed(author_1, author_2, author_3, book_1, book_2)

        Book.removed_objects.filter(id=book_1.id).restore()

        assert_is_active(author_1, book_1)
        assert_is_removed(author_2, author_3, book_2)

    def test_restore_with_related_m2m(self, make_author, book_with_category):
        book, category = book_with_category
        author_1 = book.author
        author_2, author_3 = make_author(_quantity=2)
        Author.active_objects.all().remove()
        category.remove()
        assert_is_removed(author_1, author_2, author_3, book, category)

        Author.removed_objects.filter(id=author_1.id).restore(with_related=True)

        assert_is_active(author_1, book)
        assert_is_removed(author_2, author_3, category)

    @patch.object(RemovedObjectsQuerySet, "purge")
    def test_delete(self, purge_mock, test_author):
        Author.removed_objects.all().delete()
        purge_mock.assert_called_once()

    @freeze_time("2025-05-01")
    def test_purge(self, make_author):
        author_1, author_2, author_3 = make_author(_quantity=3)
        author_1.remove()
        assert_is_active(author_2, author_3)
        assert_is_removed(author_1)

        with freeze_time("2025-05-31"):
            Author.removed_objects.all().purge()
        assert Author.objects.filter(id=author_1.id).exists()
        assert Author.objects.count() == 3

        with freeze_time("2025-06-01"):
            Author.removed_objects.all().purge()
        assert Author.objects.filter(id=author_1.id).exists() is False
        assert Author.objects.count() == 2

    @freeze_time("2025-05-01")
    def test_expired(self, make_author):
        author_1, author_2, author_3 = make_author(_quantity=3)
        author_1.remove()
        assert_is_active(author_2, author_3)
        assert_is_removed(author_1)

        assert Author.removed_objects.all().expired().count() == 0

        with freeze_time("2025-05-31"):
            assert Author.removed_objects.all().expired().count() == 0

        with freeze_time("2025-06-01"):
            assert Author.removed_objects.all().expired().count() == 1


@pytest.mark.django_db
class TestRemovedObjectsQuerySetWithAllRelations:
    def test_restore_author(self, make_author_with_all_relations):
        author_1, author_1_rels = make_author_with_all_relations()
        author_2, author_2_rels = make_author_with_all_relations()
        remove_objs(author_1, *author_1_rels, author_2, *author_2_rels)

        Author.removed_objects.filter(id=author_1.id).restore()

        assert_is_active(author_1)
        assert_is_removed(*author_1_rels, author_2, *author_2_rels)

    def test_restore_author_with_related(self, make_author_with_all_relations):
        author_1, author_1_rels = make_author_with_all_relations()
        author_2, author_2_rels = make_author_with_all_relations()
        profile, book_1, book_2, book_meta_1, book_meta_2, category_1, category_2 = (
            author_1_rels
        )
        remove_objs(author_1, *author_1_rels, author_2, *author_2_rels)

        Author.removed_objects.filter(id=author_1.id).restore(with_related=True)

        assert_is_active(author_1, profile, book_1, book_2, book_meta_1, book_meta_2)
        assert_is_removed(category_1, category_2, author_2, *author_2_rels)

    def test_restore_profile(self, make_author_with_all_relations):
        author_1, author_1_rels = make_author_with_all_relations()
        author_2, author_2_rels = make_author_with_all_relations()
        profile, book_1, book_2, book_meta_1, book_meta_2, category_1, category_2 = (
            author_1_rels
        )
        remove_objs(author_1, *author_1_rels, author_2, *author_2_rels)

        AuthorProfile.removed_objects.filter(id=profile.id).restore()

        assert_is_active(author_1, profile)
        assert_is_removed(
            book_1,
            book_2,
            book_meta_1,
            book_meta_2,
            category_1,
            category_2,
            author_2,
            *author_2_rels,
        )

    def test_restore_profile_with_related(self, make_author_with_all_relations):
        author_1, author_1_rels = make_author_with_all_relations()
        author_2, author_2_rels = make_author_with_all_relations()
        profile, book_1, book_2, book_meta_1, book_meta_2, category_1, category_2 = (
            author_1_rels
        )
        remove_objs(author_1, *author_1_rels, author_2, *author_2_rels)

        AuthorProfile.removed_objects.filter(id=profile.id).restore(with_related=True)

        assert_is_active(author_1, profile)
        assert_is_removed(
            book_1,
            book_2,
            book_meta_1,
            book_meta_2,
            category_1,
            category_2,
            author_2,
            *author_2_rels,
        )

    def test_restore_book(self, make_author_with_all_relations):
        author_1, author_1_rels = make_author_with_all_relations()
        author_2, author_2_rels = make_author_with_all_relations()
        profile, book_1, book_2, book_meta_1, book_meta_2, category_1, category_2 = (
            author_1_rels
        )
        remove_objs(author_1, *author_1_rels, author_2, *author_2_rels)

        Book.removed_objects.filter(id=book_1.id).restore()

        assert_is_active(author_1, book_1)
        assert_is_removed(
            profile,
            book_2,
            book_meta_1,
            book_meta_2,
            category_1,
            category_2,
            author_2,
            *author_2_rels,
        )

    def test_restore_book_with_related(self, make_author_with_all_relations):
        author_1, author_1_rels = make_author_with_all_relations()
        author_2, author_2_rels = make_author_with_all_relations()
        profile, book_1, book_2, book_meta_1, book_meta_2, category_1, category_2 = (
            author_1_rels
        )
        remove_objs(author_1, *author_1_rels, author_2, *author_2_rels)

        Book.removed_objects.filter(id=book_1.id).restore(with_related=True)

        assert_is_active(author_1, book_1, book_meta_1)
        assert_is_removed(
            profile,
            book_2,
            book_meta_2,
            category_1,
            category_2,
            author_2,
            *author_2_rels,
        )

    def test_restore_book_meta(self, make_author_with_all_relations):
        author_1, author_1_rels = make_author_with_all_relations()
        author_2, author_2_rels = make_author_with_all_relations()
        profile, book_1, book_2, book_meta_1, book_meta_2, category_1, category_2 = (
            author_1_rels
        )
        remove_objs(author_1, *author_1_rels, author_2, *author_2_rels)

        BookMeta.removed_objects.filter(id=book_meta_1.id).restore()

        assert_is_active(author_1, book_1, book_meta_1)
        assert_is_removed(
            profile,
            book_2,
            book_meta_2,
            category_1,
            category_2,
            author_2,
            *author_2_rels,
        )

    def test_restore_book_meta_with_related(self, make_author_with_all_relations):
        author_1, author_1_rels = make_author_with_all_relations()
        author_2, author_2_rels = make_author_with_all_relations()
        profile, book_1, book_2, book_meta_1, book_meta_2, category_1, category_2 = (
            author_1_rels
        )
        remove_objs(author_1, *author_1_rels, author_2, *author_2_rels)

        BookMeta.removed_objects.filter(id=book_meta_1.id).restore(with_related=True)

        assert_is_active(author_1, book_1, book_meta_1)
        assert_is_removed(
            profile,
            book_2,
            book_meta_2,
            category_1,
            category_2,
            author_2,
            *author_2_rels,
        )

    def test_restore_category(self, make_author_with_all_relations):
        author_1, author_1_rels = make_author_with_all_relations()
        author_2, author_2_rels = make_author_with_all_relations()
        profile, book_1, book_2, book_meta_1, book_meta_2, category_1, category_2 = (
            author_1_rels
        )
        remove_objs(author_1, *author_1_rels, author_2, *author_2_rels)

        BookCategory.removed_objects.filter(id=category_1.id).restore()

        assert_is_active(category_1)
        assert_is_removed(
            author_1,
            profile,
            book_1,
            book_2,
            book_meta_1,
            book_meta_2,
            category_2,
            author_2,
            *author_2_rels,
        )

    def test_restore_category_with_related(self, make_author_with_all_relations):
        author_1, author_1_rels = make_author_with_all_relations()
        author_2, author_2_rels = make_author_with_all_relations()
        profile, book_1, book_2, book_meta_1, book_meta_2, category_1, category_2 = (
            author_1_rels
        )
        remove_objs(author_1, *author_1_rels, author_2, *author_2_rels)

        BookCategory.removed_objects.filter(id=category_1.id).restore(with_related=True)

        assert_is_active(category_1)
        assert_is_removed(
            author_1,
            profile,
            book_1,
            book_2,
            book_meta_1,
            book_meta_2,
            category_2,
            author_2,
            *author_2_rels,
        )


@pytest.mark.django_db
class TestAllObjectsQuerySet:
    @patch.object(AllObjectsQuerySet, "remove")
    def test_delete(self, remove_mock):
        Author.objects.all().delete()
        remove_mock.assert_called_once()



import uuid

from django.db import models

from django_resurrected.models import SoftDeleteModel

from .constants import BookFormat


class BaseModel(SoftDeleteModel):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)

    class Meta:
        abstract = True


class Author(BaseModel):
    name = models.CharField(max_length=100)


class AuthorProfile(BaseModel):
    author = models.OneToOneField(
        Author, on_delete=models.CASCADE, related_name="profile"
    )
    bio = models.TextField(blank=True)


class BookCategory(BaseModel):
    name = models.CharField(max_length=100, unique=True)


class Book(BaseModel):
    title = models.CharField(max_length=100)
    author = models.ForeignKey(Author, on_delete=models.CASCADE)
    categories = models.ManyToManyField(BookCategory, related_name="books", blank=True)


class BookMeta(BaseModel):
    book = models.OneToOneField(Book, on_delete=models.CASCADE, related_name="meta")
    format = models.CharField(
        max_length=30, choices=BookFormat.choices, default=BookFormat.PAPERBACK
    )



from __future__ import annotations

from django.db import models

from django_resurrected.collectors import Collector


class BaseQuerySet(models.QuerySet):
    def active(self):
        return self.filter(is_removed=False)

    def removed(self):
        return self.filter(is_removed=True)

    def _get_collector(self) -> Collector:
        return Collector(using=self.db, origin=self)

    def hard_delete(self):
        return super().delete()


class ActiveObjectsQuerySet(BaseQuerySet):
    def remove(self):
        collector = self._get_collector()
        collector.collect_reverse_related(self)
        return collector.remove()

    def delete(self):
        return self.remove()


class RemovedObjectsQuerySet(BaseQuerySet):
    def restore(self, with_related: bool = False):
        collector = self._get_collector()
        collector.collect_forward_related(self)
        collector.collect_reverse_related(self, collect_related=with_related)
        return collector.restore()

    def expired(self):
        return self.removed().filter(removed_at__lt=self.model.retention_limit)

    def purge(self):
        return self.expired().hard_delete()  # type: ignore[attr-defined]

    def delete(self):
        return self.purge()


class AllObjectsQuerySet(ActiveObjectsQuerySet, RemovedObjectsQuerySet):
    pass


class AllObjectsManager(models.Manager):
    def get_queryset(self):
        return AllObjectsQuerySet(self.model, using=self._db)


class ActiveObjectsManager(models.Manager):
    def get_queryset(self):
        return ActiveObjectsQuerySet(self.model, using=self._db).active()


class RemovedObjectsManager(models.Manager):
    def get_queryset(self):
        return RemovedObjectsQuerySet(self.model, using=self._db).removed()



from datetime import datetime

import pytest
import pytz
from freezegun import freeze_time
from test_app.models import Author

from django_resurrected.managers import ActiveObjectsManager
from django_resurrected.managers import AllObjectsManager
from django_resurrected.managers import RemovedObjectsManager
from tests.conftest import assert_is_active
from tests.conftest import assert_is_removed
from tests.conftest import remove_objs


@pytest.mark.django_db
class TestSoftDeleteModel:
    def test_manager_type(self, test_author):
        assert isinstance(Author.objects, AllObjectsManager)
        assert isinstance(Author.active_objects, ActiveObjectsManager)
        assert isinstance(Author.removed_objects, RemovedObjectsManager)

    @freeze_time("2025-05-01")
    def test_is_expired(self, test_author, monkeypatch):
        assert_is_active(test_author)
        assert test_author.retention_days == 30
        assert test_author.is_expired is False

        test_author.remove()

        assert_is_removed(test_author)
        assert test_author.is_expired is False

        with freeze_time("2025-05-31"):
            assert test_author.is_expired is False

        with freeze_time("2025-06-01"):
            assert test_author.is_expired

            monkeypatch.setattr(Author, "retention_days", None)
            assert test_author.is_expired is False

    @freeze_time("2025-05-01")
    def test_remove(self, test_author):
        assert_is_active(test_author)

        test_author.remove()

        assert_is_removed(test_author, removed_at=datetime(2025, 5, 1, tzinfo=pytz.utc))

    @freeze_time("2025-05-01")
    def test_remove_with_relation_o2o_cascade(self, author_with_profile):
        author, profile = author_with_profile
        assert_is_active(author, profile)

        author.remove()

        assert_is_removed(
            author, profile, removed_at=datetime(2025, 5, 1, tzinfo=pytz.utc)
        )

    @freeze_time("2025-05-01")
    def test_remove_with_relation_m2o_cascade(self, author_with_books):
        author, books = author_with_books
        book1, book2 = books
        assert_is_active(author, book1, book2)

        author.remove()

        assert_is_removed(
            author, book1, book2, removed_at=datetime(2025, 5, 1, tzinfo=pytz.utc)
        )

    def test_remove_with_relation_m2m(self, book_with_category):
        book, category = book_with_category
        assert_is_active(book, category)

        book.remove()

        assert_is_removed(book)
        # NOTE: M2M relations are not removed, as Django doesn't include them in cascade
        # operations.
        assert_is_active(category)

    def test_hard_delete(self, test_author):
        assert Author.objects.filter(id=test_author.id).exists()
        test_author.hard_delete()
        assert Author.objects.filter(id=test_author.id).exists() is False

    @freeze_time("2025-05-01")
    def test_delete(self, test_author):
        assert_is_active(test_author)

        test_author.delete()

        assert_is_removed(test_author)

        with freeze_time("2025-06-01"):
            test_author.delete()
            assert Author.objects.filter(id=test_author.id).exists() is False

    def test_restore(self, test_author):
        test_author.remove()
        assert_is_removed(test_author)

        test_author.restore()

        assert_is_active(test_author)

    def test_restore_without_related_o2o_cascade(self, author_with_profile):
        author, profile = author_with_profile
        author.remove()
        assert_is_removed(author, profile)

        result = author.restore()

        assert_is_active(author)
        assert_is_removed(profile)
        assert result == (1, {"test_app.Author": 1})

    def test_restore_with_related_o2o_cascade(self, author_with_profile):
        author, profile = author_with_profile
        author.remove()
        assert_is_removed(author, profile)

        result = author.restore(with_related=True)

        assert_is_active(author, profile)
        assert result == (2, {"test_app.Author": 1, "test_app.AuthorProfile": 1})

    def test_restore_with_related_m2o_cascade(self, author_with_books):
        author, books = author_with_books
        author.remove()
        assert_is_removed(author, *books)

        author.restore(with_related=True)

        assert_is_active(author, *books)

    def test_restore_with_related_o2m(self, author_with_books):
        author, books = author_with_books
        book_1, book_2 = books
        author.remove()
        assert_is_removed(author, book_1, book_2)

        book_1.restore(with_related=True)

        assert_is_active(author, book_1)
        assert_is_removed(book_2)

    def test_restore_with_related_m2m(self, book_with_category):
        book, book_category = book_with_category
        book.remove()
        book_category.remove()
        assert_is_removed(book, book_category)

        book.restore(with_related=True)

        assert_is_active(book)
        # NOTE: M2M relations are not restored, as Django doesn't include them in
        # cascade operations.
        assert_is_removed(book_category)


@pytest.mark.django_db
class TestSoftDeleteModelWithAllRelations:
    def test_remove_author(self, make_author_with_all_relations):
        author_1, author_1_rels = make_author_with_all_relations()
        author_2, author_2_rels = make_author_with_all_relations()
        profile, book_1, book_2, book_meta_1, book_meta_2, category_1, category_2 = (
            author_1_rels
        )
        author_1.remove()

        assert_is_removed(author_1, profile, book_1, book_2, book_meta_1, book_meta_2)
        assert_is_active(category_1, category_2, author_2, *author_2_rels)

    def test_remove_profile(self, make_author_with_all_relations):
        author_1, author_1_rels = make_author_with_all_relations()
        author_2, author_2_rels = make_author_with_all_relations()
        profile, book_1, book_2, book_meta_1, book_meta_2, category_1, category_2 = (
            author_1_rels
        )

        profile.remove()

        assert_is_removed(profile)
        assert_is_active(
            author_1,
            book_1,
            book_2,
            book_meta_1,
            book_meta_2,
            category_1,
            category_2,
            author_2,
            *author_2_rels,
        )

    def test_remove_book(self, make_author_with_all_relations):
        author_1, author_1_rels = make_author_with_all_relations()
        author_2, author_2_rels = make_author_with_all_relations()
        profile, book_1, book_2, book_meta_1, book_meta_2, category_1, category_2 = (
            author_1_rels
        )

        book_1.remove()

        assert_is_removed(book_1, book_meta_1)
        assert_is_active(
            profile,
            book_2,
            book_meta_2,
            category_1,
            category_2,
            author_2,
            *author_2_rels,
        )

    def test_remove_book_meta(self, make_author_with_all_relations):
        author_1, author_1_rels = make_author_with_all_relations()
        author_2, author_2_rels = make_author_with_all_relations()
        profile, book_1, book_2, book_meta_1, book_meta_2, category_1, category_2 = (
            author_1_rels
        )

        book_meta_1.remove()

        assert_is_removed(book_meta_1)
        assert_is_active(
            author_1,
            book_1,
            profile,
            book_2,
            book_meta_2,
            category_1,
            category_2,
            author_2,
            *author_2_rels,
        )

    def test_remove_category(self, make_author_with_all_relations):
        author_1, author_1_rels = make_author_with_all_relations()
        author_2, author_2_rels = make_author_with_all_relations()
        profile, book_1, book_2, book_meta_1, book_meta_2, category_1, category_2 = (
            author_1_rels
        )

        category_1.remove()

        assert_is_removed(category_1)
        assert_is_active(
            author_1,
            profile,
            book_1,
            book_2,
            book_meta_1,
            book_meta_2,
            category_2,
            author_2,
            *author_2_rels,
        )

    def test_restore_author(self, make_author_with_all_relations):
        author_1, author_1_rels = make_author_with_all_relations()
        author_2, author_2_rels = make_author_with_all_relations()
        remove_objs(author_1, *author_1_rels, author_2, *author_2_rels)

        author_1.restore()

        assert_is_active(author_1)
        assert_is_removed(*author_1_rels, author_2, *author_2_rels)

    def test_restore_author_with_related(self, make_author_with_all_relations):
        author_1, author_1_rels = make_author_with_all_relations()
        author_2, author_2_rels = make_author_with_all_relations()
        profile, book_1, book_2, book_meta_1, book_meta_2, category_1, category_2 = (
            author_1_rels
        )
        remove_objs(author_1, *author_1_rels, author_2, *author_2_rels)

        author_1.restore(with_related=True)

        assert_is_active(author_1, profile, book_1, book_2, book_meta_1, book_meta_2)
        assert_is_removed(category_1, category_2, author_2, *author_2_rels)

    def test_restore_profile(self, make_author_with_all_relations):
        author_1, author_1_rels = make_author_with_all_relations()
        author_2, author_2_rels = make_author_with_all_relations()
        profile, book_1, book_2, book_meta_1, book_meta_2, category_1, category_2 = (
            author_1_rels
        )
        remove_objs(author_1, *author_1_rels, author_2, *author_2_rels)

        profile.restore()

        assert_is_active(author_1, profile)
        assert_is_removed(
            book_1,
            book_2,
            book_meta_1,
            book_meta_2,
            category_1,
            category_2,
            author_2,
            *author_2_rels,
        )

    def test_restore_profile_with_related(self, make_author_with_all_relations):
        author_1, author_1_rels = make_author_with_all_relations()
        author_2, author_2_rels = make_author_with_all_relations()
        profile, book_1, book_2, book_meta_1, book_meta_2, category_1, category_2 = (
            author_1_rels
        )
        remove_objs(author_1, *author_1_rels, author_2, *author_2_rels)

        profile.restore(with_related=True)

        assert_is_active(author_1, profile)
        assert_is_removed(
            book_1,
            book_2,
            book_meta_1,
            book_meta_2,
            category_1,
            category_2,
            author_2,
            *author_2_rels,
        )

    def test_restore_book(self, make_author_with_all_relations):
        author_1, author_1_rels = make_author_with_all_relations()
        author_2, author_2_rels = make_author_with_all_relations()
        profile, book_1, book_2, book_meta_1, book_meta_2, category_1, category_2 = (
            author_1_rels
        )
        remove_objs(author_1, *author_1_rels, author_2, *author_2_rels)

        book_1.restore()

        assert_is_active(author_1, book_1)
        assert_is_removed(
            profile,
            book_2,
            book_meta_1,
            book_meta_2,
            category_1,
            category_2,
            author_2,
            *author_2_rels,
        )

    def test_restore_book_with_related(self, make_author_with_all_relations):
        author_1, author_1_rels = make_author_with_all_relations()
        author_2, author_2_rels = make_author_with_all_relations()
        profile, book_1, book_2, book_meta_1, book_meta_2, category_1, category_2 = (
            author_1_rels
        )
        remove_objs(author_1, *author_1_rels, author_2, *author_2_rels)

        book_1.restore(with_related=True)

        assert_is_active(author_1, book_1, book_meta_1)
        assert_is_removed(
            profile,
            book_2,
            book_meta_2,
            category_1,
            category_2,
            author_2,
            *author_2_rels,
        )

    def test_restore_book_meta(self, make_author_with_all_relations):
        author_1, author_1_rels = make_author_with_all_relations()
        author_2, author_2_rels = make_author_with_all_relations()
        profile, book_1, book_2, book_meta_1, book_meta_2, category_1, category_2 = (
            author_1_rels
        )
        remove_objs(author_1, *author_1_rels, author_2, *author_2_rels)

        book_meta_1.restore()

        assert_is_active(author_1, book_1, book_meta_1)
        assert_is_removed(
            profile,
            book_2,
            book_meta_2,
            category_1,
            category_2,
            author_2,
            *author_2_rels,
        )

    def test_restore_book_meta_with_related(self, make_author_with_all_relations):
        author_1, author_1_rels = make_author_with_all_relations()
        author_2, author_2_rels = make_author_with_all_relations()
        profile, book_1, book_2, book_meta_1, book_meta_2, category_1, category_2 = (
            author_1_rels
        )
        remove_objs(author_1, *author_1_rels, author_2, *author_2_rels)

        book_meta_1.restore(with_related=True)

        assert_is_active(author_1, book_1, book_meta_1)
        assert_is_removed(
            profile,
            book_2,
            book_meta_2,
            category_1,
            category_2,
            author_2,
            *author_2_rels,
        )

    def test_restore_category(self, make_author_with_all_relations):
        author_1, author_1_rels = make_author_with_all_relations()
        author_2, author_2_rels = make_author_with_all_relations()
        profile, book_1, book_2, book_meta_1, book_meta_2, category_1, category_2 = (
            author_1_rels
        )
        remove_objs(author_1, *author_1_rels, author_2, *author_2_rels)

        category_1.restore()

        assert_is_active(category_1)
        assert_is_removed(
            author_1,
            profile,
            book_1,
            book_2,
            book_meta_1,
            book_meta_2,
            category_2,
            author_2,
            *author_2_rels,
        )

    def test_restore_category_with_related(self, make_author_with_all_relations):
        author_1, author_1_rels = make_author_with_all_relations()
        author_2, author_2_rels = make_author_with_all_relations()
        profile, book_1, book_2, book_meta_1, book_meta_2, category_1, category_2 = (
            author_1_rels
        )
        remove_objs(author_1, *author_1_rels, author_2, *author_2_rels)

        category_1.restore(with_related=True)

        assert_is_active(category_1)
        assert_is_removed(
            author_1,
            profile,
            book_1,
            book_2,
            book_meta_1,
            book_meta_2,
            category_2,
            author_2,
            *author_2_rels,
        )
